/*
 * Copyright (c) 2005, 2006, 2007  by KoanLogic <http://www.koanlogic.com>
 * All rights reserved.
 *
 * $Id: msgchan.c,v 1.5 2008/05/14 15:41:16 tat Exp $
 */

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <u/libu.h>
#include <klone/klone.h>
#include <klapp/msgchan.h>
#include "list.h"

#ifndef timercmp
#define timercmp(tvp, uvp, cmp) \
    ((tvp)->tv_sec cmp (uvp)->tv_sec || \
    (tvp)->tv_sec == (uvp)->tv_sec && \
    (tvp)->tv_usec cmp (uvp)->tv_usec)
#endif

enum { MAX_ID_LEN = 256 };
enum { MAX_DATA_LEN = 1024 };

enum { ID_BUFSZ = MAX_ID_LEN + 1 };
enum { DATA_BUFSZ = MAX_DATA_LEN + 1 };
enum { MSG_BUFSZ = ID_BUFSZ + DATA_BUFSZ + 16 };

/* event handler */
typedef struct handler_s
{
    msgchan_handler_t fun; /* callback function */
    void *arg;             /* user provided argument passed to the callback */
    char id[MAX_ID_LEN];   /* event id */ 
} handler_t;

/* periodic event handler */
typedef struct periodic_s
{
    msgchan_periodic_t fun; /* callback function */
    void *arg;              /* user provided argument passed to the callback */
    struct timeval tv;      /* fire next event after 'tv' time */
} periodic_t;

struct msgchan_s
{
    io_t *in;   /* msg channel input stream */
    io_t *out;  /* msg channel output stream */
    list_t *handlers;
    list_t *periodics;
    struct timeval earltv; /* earliest timeout event */
};

int msgchan_register(msgchan_t *mc, const char *id, msgchan_handler_t function,
        void *arg)
{
    handler_t *handler = NULL;

    dbg_err_if(mc == NULL);
    dbg_err_if(id == NULL);
    dbg_err_if(function == NULL);
    /* arg may be NULL */

    handler = u_zalloc(sizeof(handler_t));
    dbg_err_if(handler == NULL);

    dbg_err_if(strlcpy(handler->id, id, MAX_ID_LEN) >= MAX_ID_LEN);
    handler->fun = function;
    handler->arg = arg;

    dbg_err_if(list_add(mc->handlers, handler));

    return 0;
err:
    if(handler)
        list_del(mc->handlers, handler);
    if(handler)
        u_free(handler);
    return ~0;
}

int msgchan_unregister_periodic(msgchan_t *mc, msgchan_periodic_t function)
{
    periodic_t *pd;
    size_t n;
    
    for(n = 0; (pd = (periodic_t*)list_get_n(mc->periodics, n)) != NULL; ++n)
    {
        if(pd->fun == function)
        {
            list_del(mc->periodics, pd);
            u_free(pd);
            return 0;
        }
    }

    return ~0; /* not found */
}

int msgchan_register_periodic(msgchan_t *mc, msgchan_periodic_t function,
        void *arg)
{
    periodic_t *periodic = NULL;

    dbg_err_if(mc == NULL);
    dbg_err_if(function == NULL);
    /* arg may be NULL */

    periodic = u_zalloc(sizeof(periodic_t));
    dbg_err_if(periodic == NULL);

    periodic->fun = function;
    periodic->arg = arg;

    dbg_err_if(list_add(mc->periodics, periodic));

    return 0;
err:
    if(periodic)
        list_del(mc->periodics, periodic);
    if(periodic)
        u_free(periodic);
    return ~0;
}

static int msgchan_dispatch_message(msgchan_t *mc, const char *id, 
        const char *data)
{
    handler_t *h;
    size_t n;

    for(n = 0; (h = (handler_t*)list_get_n(mc->handlers, n)) != NULL; ++n)
    {
        if(strcmp(h->id, id) == 0)
        {
            dbg_err_if(h->fun == NULL);

            /* call the registered function */
            h->fun(mc, h->arg, id, data);

            return 0; /* msg dispatched */
        }
    }

    /* fall through */
err:
    return ~0; /* error or 'id' not registered */
}

static int msg_parse(const char *msg, char *id, char *data)
{
    const char *p;
    size_t len;

    p = strchr(msg, ':');
    dbg_err_if((len = p - msg) > MAX_ID_LEN);

    strncpy(id, msg, len);
    id[len] = 0;

    dbg_err_if((len = strlen(++p)) > MAX_DATA_LEN);

    strncpy(data, p, len);
    data[len] = 0;

    return 0;
err:
    crit("malformed message or parser error");
    return ~0;
}

static int msgchan_call_periodics(msgchan_t *mc)
{
    periodic_t *pd;
    struct timeval now;
    size_t n;

    gettimeofday(&now, NULL);

    /* wake up each sec when there are no periodics */
    mc->earltv.tv_sec = ( list_count(mc->periodics) ? INT_MAX : now.tv_sec + 1);
    mc->earltv.tv_usec = 0;

    for(n = 0; (pd = (periodic_t*)list_get_n(mc->periodics, n)) != NULL; ++n)
    {
        /* if now >= pd->tv  */
        if(timercmp(&now, &pd->tv, >= ))
        {
            /* fire the event */
            dbg_err_if(pd->fun == NULL);

            /* call the registered function */
            pd->fun(mc, pd->arg, &pd->tv);

            /* set tv to the timeval of the next time it must fire */
            pd->tv.tv_sec += now.tv_sec;
            pd->tv.tv_usec += now.tv_usec;
        }

        /* is pd->tv the earliest timeout? */
        if(timercmp(&pd->tv, &mc->earltv, <))
            mc->earltv = pd->tv;
    }

    return 0; 
err:
    return ~0; 
}

int msgchan_dispatch(msgchan_t *mc)
{
    char msg[MSG_BUFSZ], id[ID_BUFSZ], data[DATA_BUFSZ];
    struct timeval tv, now, r;
    int rc, sec, usec;

    /* call periodic functions if expired */
    crit_if(msgchan_call_periodics(mc));

    /* send the "server ready" message */
    dbg_err_if(msgchan_printf(mc, "{\"id\":\"ctrl\",\"data\":\"ready\"}") < 0);

    for(;;)
    {
        gettimeofday(&now, NULL);

        /* msgchan_read() may sleep at most 'r' till the next periodic func */
        if(timercmp(&mc->earltv, &now, >=))
            timersub(&mc->earltv, &now, &r);
        else
            timerclear(&r);

        if((rc = msgchan_read(mc, msg, sizeof(msg), &r)) > 0)
        {
            /* got a message */
            if(msg_parse(msg, id, data))
            {
                crit("malformed message");
                continue;
            }

            /* dispatch the message to the registered function */
            crit_ifm(msgchan_dispatch_message(mc, id, data), 
                    "msg id [%s] not registered", id);

            continue;

        } else if(rc == 0) {
            return 0; /* eof, exit */
        } else {
            if(errno != EAGAIN && errno != EINTR) 
                dbg_err_sif("msgchan socket error");
        }

        /* call periodic functions if expired */
        crit_if(msgchan_call_periodics(mc));
    }

    /* fall through */
err:
    return ~0;
}

/* note: this function cannot work with binary data that includes '\0' bytes */
int msgchan_read(msgchan_t *mc, char *buf, size_t sz, struct timeval *tv)
{
    ssize_t rc;
    fd_set rdset;
    int fd;

    /* get the underlaying file descriptor */
    fd = io_fd_get_d(mc->in);
    dbg_err_if(fd < 0);

    FD_ZERO(&rdset);
    FD_SET(fd, &rdset);

    dbg_err_if((rc = select(1+fd, &rdset, NULL, NULL, tv)) < 0);

    /* return -1 on timeout (same interface of non-blocking sockets) */
    if(rc == 0)
    {
        errno = EAGAIN;
        return -1; 
    }

    if(rc > 0)
    {
        /* the socket is readable, read the next message */
        if((rc = io_get_until(mc->in, '\0', buf, sz)) > 0)
        {
            /* ignore trailing '\0' sent by the XMLSocket object */
            return --rc;
        }
    }

    return rc;
err:
    return -1;
}

int msgchan_write(msgchan_t *mc, const char *buf, size_t sz)
{
    char ch = 0, rc = 0;

    /* send back the return code */
    dbg_err_if((rc = io_write(mc->out, buf, sz)) < 0);

    /* needed by Flash XMLSocket object */
    dbg_err_if((rc += io_write(mc->out, &ch, 1)) < 0);

    io_flush(mc->out);

    return rc;
err:
    return -1;
}

int msgchan_printf(msgchan_t *mc, const char *fmt, ...)
{
    va_list ap;
    char buf[4096];
    int rc;

    va_start(ap, fmt); 

    rc = vsnprintf(buf, sizeof(buf), fmt, ap);

    va_end(ap);

    crit_err_if(rc > sizeof(buf)); /* buffer too small */
        
    /* go safe... */
    buf[sizeof(buf) - 1] = 0; 

    return msgchan_write(mc, buf, rc);
err:
    return -1;
}

int msgchan_free(msgchan_t *mc)
{
    void *ptr;
    size_t n;

    dbg_err_if(mc == NULL);

    if(mc->handlers)
    {
        /* free all handler_t pointers */
        for(n = 0; (ptr = list_get_n(mc->handlers, n)) != NULL; ++n)
            u_free(ptr);

        /* free the list itself */
        list_free(mc->handlers);
    }

    if(mc->periodics)
    {
        /* free all periodic_t pointers */
        for(n = 0; (ptr = list_get_n(mc->periodics, n)) != NULL; ++n)
            u_free(ptr);

        /* free the list itself */
        list_free(mc->periodics);
    }

    u_free(mc);

    return 0;
err:
    return ~0;
}

int msgchan_create(io_t *in, io_t *out, msgchan_t **pmc)
{
    msgchan_t *mc = NULL;
    char ch;

    dbg_err_if(in == NULL);
    dbg_err_if(out == NULL);
    dbg_err_if(pmc == NULL);

    mc = u_zalloc(sizeof(msgchan_t));
    dbg_err_if(mc == NULL);

    mc->in = in;
    mc->out = out;

    mc->earltv.tv_sec = INT_MAX;

    dbg_err_if(list_create(&mc->handlers));
    dbg_err_if(list_create(&mc->periodics));

    /* this will avoid HTTP response */
    warn_err_if(io_codecs_remove(mc->out));

    /* read the trailing '\0' sent by the msgchanet object just after the 
     * HTTP request */
    io_getc(mc->in, &ch);

    *pmc = mc;

    return 0;
err:
    if(mc)
        msgchan_free(mc);
    return ~0;
}
